@@ -1,24 +1,56 @@
 # encoding: utf-8
-import PXDParser
-import Types
-import Utils
+
+__license__ = """
+
+Copyright (c) 2012-2014, Uwe Schmitt, all rights reserved.
+
+Redistribution and use in source and binary forms, with or without
+modification, are permitted provided that the following conditions are met:
+
+Redistributions of source code must retain the above copyright notice, this
+list of conditions and the following disclaimer.
+
+Redistributions in binary form must reproduce the above copyright notice, this
+list of conditions and the following disclaimer in the documentation and/or
+other materials provided with the distribution.
+
+Neither the name of the ETH Zurich nor the names of its contributors may be
+used to endorse or promote products derived from this software without specific
+prior written permission.
+
+THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
+ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
+FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
+SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
+CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+"""
+
+import autowrap.PXDParser as PXDParser
+import autowrap.Types as Types
+import autowrap.Utils as Utils
 import os
 from collections import defaultdict
-from tools import OrderKeepingDictionary
+from autowrap.tools import OrderKeepingDictionary
+
 
 
 __doc__ = """
 
-    the methods in this module take the class declarations created by
+    The methods in this module take the class declarations created by
     calling PXDParser.parse and generates a list of resolved class
     declarations.  'resolved' means that all template parameters are
     resolved and inherited methods are resolved from super classes.
 
-    some preliminaries which you should have in mind to understand the
+    Some preliminaries which you should have in mind to understand the
     code below:
 
-    in pxd files inheritance is declared with 'wrap-inherits'
-    annotations.  python class names are declared with 'wrap-instances'
+    In pxd files inheritance is declared with 'wrap-inherits'
+    annotations.  Python class names are declared with 'wrap-instances'
     annotations.
 
     eg
@@ -37,14 +69,50 @@ __doc__ = """
     In the end we get a Python class B_int_float which wraps B[int,
     float] and a Python class B_pure which wraps B[int,int].
 
-    If you wrap a C++ class without template parameters you can ommit
+    If you wrap a C++ class without template parameters you can omit
     the 'wrap-instances' annotation. In this case the name of the Python
     class is the same as the name of the C++ class.
 
+    There are a few additional hints you can give to the wrapper, for classes these are:
+        - wrap-ignore: will not create a wrapper for this class (e.g. abstract
+                       base class that needs to be known to Cython but cannot be wrapped)
+        - wrap-manual-memory: will allow the user to provide manual memory
+                              management of self.inst, therefore the class will
+                              not provide the automated __dealloc__ and inst
+                              attribute (but their presence is still expected). 
+                              This is useful if you cannot use the shared-ptr
+                              approach to store a reference to the C++ class
+                              (as with singletons for example).
+        - wrap_hash: if the produced class should be hashable, give a hint
+                     which method should be used for this. This method will be
+                     called on the C++ object and fed into the Python "hash"
+                     function. This implies the class also provides a
+                     operator== function.
+                     Note that the only requirement for a hash function is that
+                     equal objects produce equal values
+
+    Thus a class could look like this:
+
+        cdef cppclass B[U,V]:
+            # wrap-inherits:
+            #    C[U]
+            #    D
+            #
+            # wrap-instances:
+            #   B_int_float := B[int, float]
+            #   B_pure := B[int, int]
+            #
+            # wrap-manual-memory
+            #
+            # wrap-hash:
+            #   getStringId().c_str()
+
+
 """
 
 import logging as L
 
+
 class ResolvedTypeDef(object):
 
     def __init__(self, decl):
@@ -76,7 +144,21 @@ class ResolvedAttribute(object):
         self.wrap_ignore = decl.annotations.get("wrap-ignore", False)
 
 
+default_namespace = ""
+
+def get_namespace(pxd, default_namespace):
+    filehandle = open(pxd)
+    fulltext = filehandle.read()
+    filehandle.close()
+    import re
+    match = re.search(r"cdef extern.*?namespace\s*\"([^\"]*)\"", fulltext)
+    if not match:
+        return default_namespace
+    else:
+        return match.group(1)
+
 class ResolvedClass(object):
+
     """ contains all info for generating wrapping code of
         resolved class.
         "Resolved" means that template parameters and typedefs are resolved.
@@ -91,8 +173,18 @@ class ResolvedClass(object):
         self.attributes = attributes
 
         self.cpp_decl = decl
-        #self.items = getattr(decl, "items", [])
+        self.ns = get_namespace(decl.pxd_path, default_namespace)
+
         self.wrap_ignore = decl.annotations.get("wrap-ignore", False)
+        self.no_pxd_import = decl.annotations.get("no-pxd-import", False)
+        self.wrap_manual_memory = decl.annotations.get("wrap-manual-memory", [])
+        # fix previous code where we had a bool ... 
+        if self.wrap_manual_memory == True:
+            self.wrap_manual_memory = ["__old-model"]
+        else:
+            self.wrap_manual_memory == []
+        assert( isinstance(self.wrap_manual_memory, list) )
+        self.wrap_hash = decl.annotations.get("wrap-hash", [])
         self.local_map = local_map
         self.instance_map = instance_map
 
@@ -117,6 +209,7 @@ class ResolvedMethod(object):
         self.arguments = arguments
         self.cpp_decl = decl
         self.wrap_ignore = decl.annotations.get("wrap-ignore", False)
+        self.with_nogil = decl.annotations.get("wrap-with-no-gil", False)
         self.local_map = local_map
         self.instance_amp = instance_map
 
@@ -130,15 +223,47 @@ class ResolvedFunction(ResolvedMethod):
     pass
 
 
-
-def resolve_decls_from_files(pathes, root):
+def resolve_decls_from_files_single_thread(pathes, root):
     decls = []
-    for path in pathes:
+    for k, path in enumerate(pathes):
         full_path = os.path.join(root, path)
-        L.info("parse %s" % full_path)
+        if k % 50 == 0: 
+            L.log(25, "parsing progress %s out of %s" % (k, len(pathes)))
         decls.extend(PXDParser.parse_pxd_file(full_path))
     return _resolve_decls(decls)
 
+def resolve_decls_from_files(pathes, root, num_processes = 1):
+    if num_processes > 1:
+        return resolve_decls_from_files_multi_thread(pathes, root, num_processes)
+    else:
+        return resolve_decls_from_files_single_thread(pathes, root)
+
+def resolve_decls_from_files_multi_thread(pathes, root, num_processes):
+    """Perform parsing with multiple threads
+
+    This function distributes the work on `num_processes` processes and each
+    process works on 10 files at a time until there are no more files left to
+    work on.
+    """
+    import multiprocessing as mp
+
+    CONCURRENT_FILES_PER_CORE = 10
+    pool = mp.Pool(processes=num_processes)
+    full_pathes = [os.path.join(root, path) for path in pathes]
+
+    decls = []
+    while len(full_pathes) > 0:
+        n_work = len(full_pathes)
+        remaining = max(0, n_work - num_processes * CONCURRENT_FILES_PER_CORE)
+        args = [full_pathes.pop() for k in range(n_work - remaining)]
+        L.log(25, "parsing progress %s out of %s with %s processes" % (len(pathes)-remaining, len(pathes), num_processes))
+
+        res = pool.map(PXDParser.parse_pxd_file, args)
+        for r in res:
+            decls.extend(r)
+
+    return _resolve_decls(decls)  
+
 
 def resolve_decls_from_string(pxd_in_a_string):
     return _resolve_decls(PXDParser.parse_str(pxd_in_a_string))
@@ -162,10 +287,10 @@ def _resolve_decls(decls):
     def filter_out(tt):
         return [d for d in decls if isinstance(d, tt) and d.name is not None]
 
-    typedef_decls  = filter_out(PXDParser.CTypeDefDecl)
+    typedef_decls = filter_out(PXDParser.CTypeDefDecl)
     function_decls = filter_out(PXDParser.CppMethodOrFunctionDecl)
-    enum_decls     = filter_out(PXDParser.EnumDecl)
-    class_decls    = filter_out(PXDParser.CppClassDecl)
+    enum_decls = filter_out(PXDParser.EnumDecl)
+    class_decls = filter_out(PXDParser.CppClassDecl)
 
     class_decls = _resolve_all_inheritances(class_decls)
 
@@ -179,29 +304,24 @@ def _resolve_decls(decls):
     # register class aliases
     instance_mapping = _parse_all_wrap_instances_comments(class_decls)
     # remove local targ mappings:
-    instance_mapping = dict( (k, v0) for (k, (v0,v1)) in
-                                                      instance_mapping.items())
+    instance_mapping = dict((k, v0) for (k, (v0, v1)) in instance_mapping.items())
 
     # resolve  typedefs in class aliase values
     instance_mapping = dict((k, v.transformed(typedef_mapping)) for (k, v) in
-                                                      instance_mapping.items())
+                            instance_mapping.items())
 
     # add enum mapping to class instances mapping
     intersecting_names = set(instance_mapping) & set(enum_mapping)
-    assert not intersecting_names, "enum names and class decls overlap: %s"\
-                                   % intersecting_names
+    assert not intersecting_names, "enum names and class decls overlap: %s" % intersecting_names
 
     instance_mapping.update(enum_mapping)
 
-    functions = [_resolve_function(f, instance_mapping, typedef_mapping)
-                                                       for f in function_decls]
+    functions = [_resolve_function(f, instance_mapping, typedef_mapping) for f in function_decls]
 
     enums = [ResolvedEnum(e) for e in enum_decls]
     typedefs = [ResolvedTypeDef(t) for t in typedef_decls]
 
-    classes = _resolve_class_decls(class_decls,
-                                   typedef_mapping,
-                                   instance_mapping)
+    classes = _resolve_class_decls(class_decls, typedef_mapping, instance_mapping)
 
     return classes + enums + functions + typedefs, instance_mapping
 
@@ -289,20 +409,19 @@ def _add_inherited_methods(cdcl, super_cld, used_parameters):
 
     # check if parmetirization signature matches:
     if len(used_parameters) != len(super_targs):
-        raise Exception("deriving %s from %s does not match"
-                        % (cdcl.name, super_cld.name))
+        raise Exception("deriving %s from %s does not match" % (cdcl.name, super_cld.name))
 
     # map template parameters in super class to the parameters used in current
     # class:
     mapping = dict(zip(super_targs, used_parameters))
     # get copy of methods from super class ans transform template params:
     transformed_methods = super_cld.get_transformed_methods(mapping)
-    transformed_methods = dict((k,v) for (k,v) in transformed_methods.items()
-                                if k != super_cld.name) # remove constructors
+    transformed_methods = dict((k, v) for (k, v) in transformed_methods.items()
+                               if k != super_cld.name)  # remove constructors
     for method in transformed_methods:
         L.info("attach to %s: %s" % (cdcl.name, method))
     cdcl.attach_base_methods(transformed_methods)
-    L.info("")
+    # L.info("")
 
 
 def _build_typedef_mapping(decls):
@@ -315,7 +434,7 @@ def _build_typedef_mapping(decls):
 def _check_typedefs(decls):
     left_sides = [decl.name for decl in decls]
     if len(left_sides) != len(set(left_sides)):
-        multiples = [ls for ls in left_sides if left_sides.count(ls)>1]
+        multiples = [ls for ls in left_sides if left_sides.count(ls) > 1]
         msg = "multiple typedefs for name(s) '%s'" % (", ".join(multiples))
         raise Exception(msg)
 
@@ -387,9 +506,7 @@ def _resolve_class_decls(class_decls, typedef_mapping, instance_mapping):
     """
     all_resolved_classes = []
     for class_decl in class_decls:
-        resolved_classes = _resolve_class_decl(class_decl,
-                                               typedef_mapping,
-                                               instance_mapping)
+        resolved_classes = _resolve_class_decl(class_decl, typedef_mapping, instance_mapping)
         all_resolved_classes.extend(resolved_classes)
     return all_resolved_classes
 
@@ -406,8 +523,7 @@ def _resolve_class_decl(class_decl, typedef_mapping, i_mapping):
 
         r_attributes = []
         for adcl in class_decl.attributes:
-            r_attributes.append(_resolve_attribute(adcl, i_mapping,
-                                                                local_mapping))
+            r_attributes.append(_resolve_attribute(adcl, i_mapping, local_mapping))
 
         r_methods = []
         for (mname, mdcls) in class_decl.methods.items():
@@ -416,21 +532,19 @@ def _resolve_class_decl(class_decl, typedef_mapping, i_mapping):
                 if ignore:
                     continue
                 if mdcl.name == class_decl.name:
-                    r_method = _resolve_constructor(cinst_name, mdcl,
-                                                      i_mapping, local_mapping)
+                    r_method = _resolve_constructor(cinst_name, mdcl, i_mapping, local_mapping)
                 else:
                     r_method = _resolve_method(mdcl, i_mapping, local_mapping)
                 r_methods.append(r_method)
         r_class = ResolvedClass(cinst_name, r_methods, r_attributes,
-                class_decl, i_mapping, local_mapping)
+                                class_decl, i_mapping, local_mapping)
         resolved_classes.append(r_class)
     return resolved_classes
 
 
 def _build_local_typemap(t_param_mapping, typedef_mapping):
     # for resolving typedefed types in template instance args:
-    local_map = dict((n, t.transformed(typedef_mapping)) for (n, t) in \
-            t_param_mapping.items())
+    local_map = dict((n, t.transformed(typedef_mapping)) for (n, t) in t_param_mapping.items())
 
     # for resolving 'free' typedefs in method args and result types:
     if set(local_map) & set(typedef_mapping):
@@ -446,8 +560,8 @@ def _resolve_constructor(cinst_name, method_decl, instance_mapping, local_type_m
     result = _resolve_method_or_function(method_decl, instance_mapping,
                                          local_type_map, ResolvedMethod)
     result.name = cinst_name
-    L.info("result             : '%s'" % result)
-    L.info("")
+    # L.info("result             : '%s'" % result)
+    # L.info("")
     return result
 
 
@@ -455,8 +569,8 @@ def _resolve_method(method_decl, instance_mapping, local_type_map):
     L.info("resolve method decl: '%s'" % method_decl)
     result = _resolve_method_or_function(method_decl, instance_mapping,
                                          local_type_map, ResolvedMethod)
-    L.info("result             : '%s'" % result)
-    L.info("")
+    # L.info("result             : '%s'" % result)
+    # L.info("")
     return result
 
 
@@ -464,8 +578,8 @@ def _resolve_function(method_decl, instance_mapping, local_type_map):
     L.info("resolve function decl: '%s'" % method_decl)
     result = _resolve_method_or_function(method_decl, instance_mapping,
                                          local_type_map, ResolvedFunction)
-    L.info("result               : '%s'" % result)
-    L.info("")
+    # L.info("result               : '%s'" % result)
+    # L.info("")
     return result
 
 
@@ -475,14 +589,13 @@ def _resolve_method_or_function(method_decl, instance_mapping, local_type_map,
     resolves aliases in return and argument types
     """
     result_type = _resolve_alias(method_decl.result_type, instance_mapping,
-                                local_type_map)
+                                 local_type_map)
     args = []
     for arg_name, arg_type in method_decl.arguments:
         arg_type = _resolve_alias(arg_type, instance_mapping, local_type_map)
         args.append((arg_name, arg_type))
     name = method_decl.annotations.get("wrap-as", method_decl.name)
-    return clz(name, result_type, args, method_decl, instance_mapping,
-               local_type_map)
+    return clz(name, result_type, args, method_decl, instance_mapping, local_type_map)
 
 
 def _resolve_attribute(adecl, instance_mapping, type_map):
