@@ -1,4 +1,4 @@
-from IntHolderConverter import IntHolderConverter
+from .IntHolderConverter import IntHolderConverter
 
 def register_converters():
     from autowrap.ConversionProvider import  special_converters
